import java.io.*;
import java.util.*;
/**
* Parse an input stream containing an FM language source file
* and optionally generate postscript code to implement the movie.
* Implements a recursive descent predictive parser according
* to the FM language grammar.
* <br>$Id$
*/
public class Parser {
	/**
	* Construct a new Parser object.  Remember the CompilerIO object
	* so that we can use it for output in the parser.  Remember the 
	* SymbolTable object so that we can use it to store information
	* about program symbols.  Remember the Scanner object that supplies
	* us with a Token stream to parse.
	* Initialize the HashMap firstSets to hold int[] objects containing
	* the Token identifier values for the FIRST sets of various non-terminals.
	* The name of the associated non-terminal is used as the key.
	* The pageNumber variable is declared and initialized to 0 in the 
	* SymbolTable.
	* @param io the CompilerIO object to use in reading and writing files.
	* @param t the SymbolTable manager.  Table 0 is the reserved words
	* table, table 1 is used for program symbols.
	* @param s the lexical scanner that actually reads the input file
	*/
	public Parser(CompilerIO io, SymbolTable t, Scanner s) {
		cio = io;
		scan = s;
		symbolTable = t;
		
		Symbol pageNumber = new Symbol("pageNumber",Symbol.VARIABLE);
		pageNumber.putAttribute("value",new Integer(0));
		symbolTable.putSymbol(SC_VARIABLE,"pageNumber",pageNumber);
		
		firstSets = new HashMap();
		firstSets.put("prologBlock",new int[] {Token.KW_PROLOG});
		firstSets.put("prologStatement",new int[] {Token.ID});
		firstSets.put("pageBlocksTail",new int[] {Token.KW_SHOW});
		firstSets.put("pageBlock",new int[] {Token.KW_SHOW});
		firstSets.put("pageStatement",new int[] {Token.LCURLY,Token.KW_IF,Token.ID});
		firstSets.put("exprTail",new int[] {Token.OP_ADD,Token.OP_SUB});
		firstSets.put("termTail",new int[] {Token.OP_MUL,Token.OP_DIV});
		firstSets.put("factor",new int[] {Token.INTEGER,Token.REAL,Token.LPAREN,Token.ID});
		firstSets.put("expr",new int[] {Token.INTEGER,Token.REAL,Token.LPAREN,Token.ID});
		firstSets.put("callEnd",new int[] {Token.DOT, Token.LPAREN});
		firstSets.put("assignEnd",new int[] {Token.OP_ASSIGN});
	}
	/**
	* Starting with the first Token in the input stream, 
	* try to derive a parse tree for the entire program.
	* @return true if parsed correctly, false if error
	*/
	public boolean parse() {
		status = true;
		prevToken = null;
		theToken = scan.nextToken();
		parseProgram();
		return status;
	}
	/**
	* Parse non-terminal: program.
	* <pre>
	* program ::= movie name { movieBody } EOF
	* </pre>
	*/
	private void parseProgram() {
		traceEntry();
		try {
			matchToken(Token.KW_MOVIE);
			matchToken(Token.ID);
			matchToken(Token.LCURLY);
//			parseMovieBody();
			matchToken(Token.RCURLY);
			matchToken(Token.EOF);
		}
		catch (SyntaxException e) {
			processSyntaxException(e);
		}
		traceExit();
	}
	/**
	* Parse non-terminal: movieBody.
	* <pre>
	* movieBody ::= prologBlock pageBlocks
	* movieBody ::= pageBlocks
	* </pre>
	*/
	private void parseMovieBody() {
//		traceEntry();
//		if (isFirst(theToken,"prologBlock")) {
//			parsePrologBlock();
//		}
//		parsePageBlocks();
//		traceExit();
	}

  //TODO  lots of parse methods go here ...


	/**
	* Check to see if tokens of the given type are in the FIRST set
	* of the given non-terminal.
	* @param t the Token to check to see if its type is in the FIRST set
	* @param nonterm the String name of the non-terminal whose FIRST set
	* is to be checked
	* @return true if the given Token can be the first Token in a string
	* generated by this non-terminal, false if it cannot be the first
	* Token.
	*/
	private boolean isFirst(Token t,String nonterm) {
		int[] first = (int[])firstSets.get(nonterm);
		if (first == null) return false;
		for (int i=0; i<first.length; i++) {
			if (first[i]==t.getType()) return true;
		}
		return false;
	}
	/**
	* Check that the current Token is of the expected
	* type, and advance past it if it is okay.
	* @param type the type of the Token, as defined by the class 
	* variables in Token.
	* @throws SyntaxException if theToken is not of the expected
	* type.
	*/
	private void matchToken(int type) throws SyntaxException {
		int t = theToken.getType();
		if (t == type) {
			prevToken = theToken;
			theToken = scan.nextToken();
		} else {
			throw new SyntaxException("Expected "+Token.typeLabel(type)+
			", found "+Token.typeLabel(t));
		}
	}
	/**
	* Check that the current Token is one of the expected
	* types, and advance past it if it is okay.
	* @param type an array of valid Token types, as defined
	* by the class variables in Token.
	* @throws SyntaxException if theToken is not one of the expected
	* types.
	*/
	private void matchTokenArray(int[] type) throws SyntaxException {
		int t = theToken.getType();
		if (type != null) {
			for (int i=0; i<type.length; i++) {
				if (t==type[i]) {
					prevToken = theToken;
					theToken = scan.nextToken();
					return;
				}
			}
		}
		throw new SyntaxException("Unexpected token type: "
		+Token.typeLabel(t));
	}
	/**
	* Handle the reporting and processing required by a parse error.
	*
	* This method uses the Throwable method getStackTrace to get
	* information about the call stack and extracts the name of
	* the calling method.  This capability is not available in
	* versions of Java prior to 1.4.  
	* @param e the SyntaxException
	*/
	private void processSyntaxException(SyntaxException e) {
		StackTraceElement[] st = e.getStackTrace();
		String msg = "Parse error: "+e.getMessage()+".  Reported by "+st[1].getMethodName()+".";
		System.out.println(msg);
		cio.emitWithPrefix(msg);
		status = false;
		prevToken = theToken;
		theToken = scan.nextToken();
	}
	/**
	* Handle the reporting and processing required by an IO error.
	*
	* This method uses the Throwable method getStackTrace to get
	* information about the call stack and extracts the name of
	* the calling method.  This capability is not available in
	* versions of Java prior to 1.4.  
	* @param e the SyntaxException
	*/
	private void processIOException(IOException e) {
		StackTraceElement[] st = e.getStackTrace();
		String msg = "IO error: "+e.getMessage()+".  Reported by "+st[1].getMethodName()+".";
		System.out.println(msg);
		cio.emitWithPrefix(msg);
		status = false;
	}
	/**
	* Mutator method to set the state of method entry/exit tracing.
	* @param b true if trace output is desired, false if not
	*/
	public void setShowMethods(boolean b) {
		showMethods = b;
	}
	/**
	* Mutator method to set the state of Symbol entry display.
	* @param b true if Symbol output is desired, false if not
	*/
	public void setShowSymbols(boolean b) {
		showSymbols = b;
	}
	/**
	* Mutator method to set the state of code generation
	* @param b true if code should be generated, false if not.
	*/
	public void setShowCode(boolean b) {
		showCode = b;
	}
	/**
	* Optionally print a message that we are entering a method.
	*
	* This method uses the Throwable method getStackTrace to get
	* information about the call stack and extracts the name of
	* the calling method.  This capability is not available in
	* versions of Java prior to 1.4.  This method assumes that
	* there is an instance variable named "tracer" that has been
	* initialized with a reference to a Throwable object.
	*/
	private void traceEntry() {
		if (showMethods) {
			tracer.fillInStackTrace();
			StackTraceElement[] st = tracer.getStackTrace();
			traceEntry(st[1].getMethodName());
		}
	}
	/**
	* Optionally print a message that we are entering a method.
	* @param s the name of the method we are entering
	*/
	private void traceEntry(String s) {
		if (showMethods) {
			String msg = "Enter: "+s;
			System.out.println(msg);
			cio.emitWithPrefix(msg);
		}
	}
	/**
	* Optionally print a message that we are exiting a method.
	*
	* This method uses the Throwable method getStackTrace to get
	* information about the call stack and extracts the name of
	* the calling method.  This capability is not available in
	* versions of Java prior to 1.4.  This method assumes that
	* there is an instance variable named "tracer" that has been
	* initialized with a reference to a Throwable object.
	*/
	private void traceExit() {
		if (showMethods) {
			tracer.fillInStackTrace();
			StackTraceElement[] st = tracer.getStackTrace();
			traceExit(st[1].getMethodName());
		}
	}
	/**
	* Optionally print a message that we are exiting a method.
	* @param s the name of the method we are exiting
	*/
	private void traceExit(String s) {
		if (showMethods) {
			String msg = "Exit:  "+s;
			System.out.println(msg);
			cio.emitWithPrefix(msg);
		}
	}
	/**
	* Optionally display a Symbol
	* @param sym the Symbol to display
	*/
	private void traceSymbol(Symbol sym) {
		if (showSymbols) {
			String msg = "Symbol: "+sym.toString();
			System.out.println(msg);
			cio.emitWithPrefix(msg);
		}
	}
	//--------------------------------------------------------
	/**
	* theToken is always the Token that we should look at next.
	* When a parse method returns, theToken contains a reference
	* to the token that follows the end of the non-terminal that
	* was just parsed.
	*/
	private Token theToken;
	/** prevToken is always the last Token that has been matched.
	* This means that when matchToken succeeds, prevToken contains
	* a reference to the Token that was just matched.
	*/
	private Token prevToken;
	/**
	* The CompilerIO object that we can use to write on the
	* output stream if needed.
	*/
	private CompilerIO cio;
	/**
	* The Scanner that actually reads the input stream and 
	* reports the tokens that it finds.
	*/
	private Scanner scan;
	/**
	* This status variable is set to true before the parse 
	* begins.  If any of the parse methods encounter a problem,
	* they set this variable to false.  After the parse is
	* complete, the status is returned to the caller.
	*/
	private boolean status;
	/**
	* This Throwable is used to get stack trace information
	*/
	private Throwable tracer = new Throwable();
	/**
	* Enable / disable method entry/exit tracing.
	*/
	private boolean showMethods = false;
	/**
	* Enable / disable Symbol entry display.
	*/
	private boolean showSymbols = false;
	/**
	* Enable / disable code generation.
	*/
	private boolean showCode = false;
	/**
	* HashMap containing int arrays of Token types, keyed by the
	* non-terminal whose FIRST set each array represents.
	*/
	private HashMap firstSets;
	/**
	* Symbol table manager.
	*/
	private SymbolTable symbolTable;
	
	/** 
	* Define the symbol tables that we use in the parser.
	* This number is the symbol table index.
	*/
	private static final int SC_KEYWORD = 0;
	private static final int SC_VARIABLE = 1;
}

